@@ -162,7 +162,11 @@ pub enum StakePoolInstruction {
     /// 10. `[]` Stake Config sysvar
     /// 11. `[]` System program
     /// 12. `[]` Stake program
-    ///  userdata: amount of lamports to split into the transient stake account
+    ///  userdata: amount of lamports to increase on the given validator.
+    ///  The actual amount split into the transient stake account is:
+    ///  `lamports + stake_rent_exemption`
+    ///  The rent-exemption of the stake account is withdrawn back to the reserve
+    ///  after it is merged.
     IncreaseValidatorStake(u64),
 
     /// (Staker only) Set the preferred deposit or withdraw stake account for the
@@ -172,9 +176,9 @@ pub enum StakePoolInstruction {
     /// between SOL staked on different validators, the staker can force all
     /// deposits and/or withdraws to go to one chosen account, or unset that account.
     ///
-    /// 0. `[]` Stake pool
+    /// 0. `[w]` Stake pool
     /// 1. `[s]` Stake pool staker
-    /// 2. `[w]` Validator list
+    /// 2. `[]` Validator list
     ///
     /// Fails if the validator is not part of the stake pool.
     SetPreferredValidator {
@@ -227,6 +231,12 @@ pub enum StakePoolInstruction {
     ///   7. `[]` Pool token program
     UpdateStakePoolBalance,
 
+    ///   Cleans up validator stake account entries marked as `ReadyForRemoval`
+    ///
+    ///   0. `[]` Stake pool
+    ///   1. `[w]` Validator stake list storage account
+    CleanupRemovedValidatorEntries,
+
     ///   Deposit some stake into the pool.  The output is a "pool" token representing ownership
     ///   into the pool. Inputs are converted to the current ratio.
     ///
@@ -511,9 +521,9 @@ pub fn set_preferred_validator(
     Instruction {
         program_id: *program_id,
         accounts: vec![
-            AccountMeta::new_readonly(*stake_pool_address, false),
+            AccountMeta::new(*stake_pool_address, false),
             AccountMeta::new_readonly(*staker, true),
-            AccountMeta::new(*validator_list_address, false),
+            AccountMeta::new_readonly(*validator_list_address, false),
         ],
         data: StakePoolInstruction::SetPreferredValidator {
             validator_type,
@@ -726,6 +736,25 @@ pub fn update_stake_pool_balance(
     }
 }
 
+/// Creates `CleanupRemovedValidatorEntries` instruction (removes entries from the validator list)
+pub fn cleanup_removed_validator_entries(
+    program_id: &Pubkey,
+    stake_pool: &Pubkey,
+    validator_list_storage: &Pubkey,
+) -> Instruction {
+    let accounts = vec![
+        AccountMeta::new_readonly(*stake_pool, false),
+        AccountMeta::new(*validator_list_storage, false),
+    ];
+    Instruction {
+        program_id: *program_id,
+        accounts,
+        data: StakePoolInstruction::CleanupRemovedValidatorEntries
+            .try_to_vec()
+            .unwrap(),
+    }
+}
+
 /// Creates all `UpdateValidatorListBalance` and `UpdateStakePoolBalance`
 /// instructions for fully updating a stake pool each epoch
 pub fn update_stake_pool(
@@ -734,7 +763,7 @@ pub fn update_stake_pool(
     validator_list: &ValidatorList,
     stake_pool_address: &Pubkey,
     no_merge: bool,
-) -> Vec<Instruction> {
+) -> (Vec<Instruction>, Vec<Instruction>) {
     let vote_accounts: Vec<Pubkey> = validator_list
         .validators
         .iter()
@@ -744,10 +773,10 @@ pub fn update_stake_pool(
     let (withdraw_authority, _) =
         find_withdraw_authority_program_address(program_id, stake_pool_address);
 
-    let mut instructions: Vec<Instruction> = vec![];
+    let mut update_list_instructions: Vec<Instruction> = vec![];
     let mut start_index = 0;
     for accounts_chunk in vote_accounts.chunks(MAX_VALIDATORS_TO_UPDATE) {
-        instructions.push(update_validator_list_balance(
+        update_list_instructions.push(update_validator_list_balance(
             program_id,
             stake_pool_address,
             &withdraw_authority,
@@ -760,16 +789,23 @@ pub fn update_stake_pool(
         start_index += MAX_VALIDATORS_TO_UPDATE as u32;
     }
 
-    instructions.push(update_stake_pool_balance(
-        program_id,
-        stake_pool_address,
-        &withdraw_authority,
-        &stake_pool.validator_list,
-        &stake_pool.reserve_stake,
-        &stake_pool.manager_fee_account,
-        &stake_pool.pool_mint,
-    ));
-    instructions
+    let final_instructions = vec![
+        update_stake_pool_balance(
+            program_id,
+            stake_pool_address,
+            &withdraw_authority,
+            &stake_pool.validator_list,
+            &stake_pool.reserve_stake,
+            &stake_pool.manager_fee_account,
+            &stake_pool.pool_mint,
+        ),
+        cleanup_removed_validator_entries(
+            program_id,
+            stake_pool_address,
+            &stake_pool.validator_list,
+        ),
+    ];
+    (update_list_instructions, final_instructions)
 }
 
 /// Creates instructions required to deposit into a stake pool, given a stake
